import os

import numpy as np
import pytest
import unyt
import yt

from yt_aspect.data_structures import _is_aspect_datafile, _is_pvtu_datafile


def get_file_path_from_data_info(dataset_type, data_info_dict, dataset_name):
    ds_info = data_info_dict[dataset_type][dataset_name]
    final_folder = os.path.split(ds_info["archive_path"])[-1]
    fi_dir = os.path.join(
        data_info_dict["base_dir"],
        ds_info["archive_path"],
        final_folder,
        ds_info["relative_unpacked_path"],
    )
    full_file = os.path.join(fi_dir, ds_info["sample_file"])
    return full_file


@pytest.mark.parametrize(
    "dataset_type, dataset_name, expected_dims",
    (
        ("ASPECT", "cartesian_3D_nproc4", 3),
        ("ASPECT", "cartesian_3D_nproc1", 3),
        ("ASPECT", "shell_2D", 2),
        ("PVTU", "cleaned_aspect", 3),
    ),
)
def test_3d_aspect_load(
    pvtu_test_data, dataset_type, dataset_name, expected_dims, tmp_path
):

    fi = get_file_path_from_data_info(dataset_type, pvtu_test_data, dataset_name)

    if os.path.isfile(fi) is False:
        pytest.skip(f"Could not locate {fi}")

    ds = yt.load(fi)
    assert ds.dataset_type == dataset_type.lower()
    assert ds.dimensionality == expected_dims

    ad = ds.all_data()
    T = ad[("connect0", "T")][0]
    if dataset_type == "ASPECT":
        assert T.units == unyt.K

    _ = ad[("connect0", "velocity_x")][0]

    outfi = tmp_path / "test.png"
    slc = yt.SlicePlot(ds, "z", ("connect0", "T"))
    slc.save(outfi)
    assert os.path.isfile(outfi)


def _get_slice_frb(ds):
    slc = yt.SlicePlot(ds, "z", ("connect0", "Temperature2"))
    slc._setup_plots()
    return slc.frb[("connect0", "Temperature2")]


def test_element_validation(pvtu_test_data):

    # this sample ds does not include null elements... so not a great test.
    fi = get_file_path_from_data_info("PVTU", pvtu_test_data, "two2_with_invalid_els")
    ds_1 = yt.load(fi, detect_null_elements=True)
    frb_1 = _get_slice_frb(ds_1)
    n_finite_no_null = np.isfinite(frb_1).sum()

    ds_0 = yt.load(fi, detect_null_elements=False)
    frb_0 = _get_slice_frb(ds_0)
    n_finite_w_null = np.isfinite(frb_0).sum()

    assert n_finite_no_null > n_finite_w_null


@pytest.mark.parametrize(
    "dataset_type, dataset_name",
    (
        ("ASPECT", "cartesian_3D_nproc4"),
        ("ASPECT", "cartesian_3D_nproc1"),
        ("ASPECT", "shell_2D"),
        ("PVTU", "cleaned_aspect"),
    ),
)
def test_dataset_validation(pvtu_test_data, dataset_type, dataset_name):
    fi = get_file_path_from_data_info(dataset_type, pvtu_test_data, dataset_name)
    if dataset_type == "ASPECT":
        assert _is_aspect_datafile(fi)
        assert _is_pvtu_datafile(fi)
    else:
        assert _is_aspect_datafile(fi) is False
        assert _is_pvtu_datafile(fi)


def test_dataset_invalidation():
    assert _is_aspect_datafile("file/that/does/not/exist") is False
    assert _is_pvtu_datafile("file/that/does/not/exist") is False

    assert _is_aspect_datafile("file/that/looks/like/file.pvtu") is False
    assert _is_pvtu_datafile("file/that/looks/like/file.pvtu") is False



from typing import Union

from packaging.version import Version
from yt import __version__ as yt_version_string

yt_version = Version(yt_version_string)


def yt_is(comparison: str, test_version: Union[Version, str]) -> bool:
    # check if yt version is less than
    if isinstance(test_version, str):
        test_version = Version(test_version)

    if comparison == "<":
        return yt_version < test_version
    elif comparison == "<=":
        return yt_version <= test_version
    if comparison == ">":
        return yt_version > test_version
    elif comparison == ">=":
        return yt_version >= test_version
    if comparison == "!=":
        return yt_version != test_version
    elif comparison == "==":
        return yt_version == test_version
    else:
        raise ValueError("comparison string must be one of: <, <=, >, >=, !=, ==")



import numpy as np
import pytest
from packaging.version import Version

from yt_aspect._utilities import yt_is, yt_version, yt_version_string
from yt_aspect.util import _valid_element_mask


@pytest.mark.parametrize("compare_str", (">", ">=", "!="))
def test_yt_version_comparison_gt(compare_str):
    assert yt_is(compare_str, "1.0.0")
    assert yt_is(compare_str, Version("1.0.0"))


@pytest.mark.parametrize("compare_str", ("<", "<="))
def test_yt_version_comparison_lt(compare_str):
    assert yt_is(compare_str, "10000.0.0")
    assert yt_is(compare_str, Version("10000.0.0"))


def test_yt_version_eq():
    assert yt_is("==", yt_version)
    assert yt_is("==", yt_version_string)


def test_yt_version_bad():
    with pytest.raises(ValueError, match="comparison string must be one of:"):
        _ = yt_is("~", Version("1.0.0"))


@pytest.mark.parametrize(
    "n_elements,nodes_per_element,n_invalid", [(5, 3, 2), (5, 8, 2), (5, 8, 0)]
)
def test_valid_element_mask(n_elements, nodes_per_element, n_invalid):

    one_element = np.arange(nodes_per_element)
    test_conn = np.repeat(one_element[:, np.newaxis], n_elements, axis=1).transpose()

    if n_invalid > 0:
        test_conn[-n_invalid:, :] = 1

    valid_elements, n_actual = _valid_element_mask(test_conn)
    assert valid_elements.sum() == n_elements - n_invalid
    assert n_actual == n_invalid
